"""
Multi-indexing tools after Bertran's notation

Polynomials can either be indexed with single index or multi-index
Bertran is a notation for switching between the two.
"""

import chaospy.bertran.operators
import chaospy.bertran.fourier
import chaospy.bertran.sparse

from chaospy.bertran.operators import *
from chaospy.bertran.sparse import sparse_segment
