"""
Collection of mathematical operators.
"""
from .addition import add
from .logarithm import log
from .logarithm10 import log10
from .logarithmn import logn
from .multiply import mul
from .negative import neg
from .power import pow
from .trunkation import trunk

from .trignometric import (
    Sin, Arcsin, Cos, Arccos, Tan, Arctan,
)
from .hyperbolic import (
    Sinh, Arcsinh, Cosh, Arccosh, Tanh, Arctanh
)
